"""
Exports WIDE_EASTASIAN table keyed by supporting unicode version level.

This code generated by python wcwidth project.
"""
# pylint: disable=duplicate-code
WIDE_EASTASIAN = {
    '17.0.0': (
        # Source: EastAsianWidth-17.0.0.txt
        # Date: 2025-07-24, 00:12:54 GMT
        #
        (0x01100, 0x0115f,),  # Hangul Choseong Kiyeok  ..Hangul Choseong Filler
        (0x0231a, 0x0231b,),  # Watch                   ..Hourglass
        (0x02329, 0x0232a,),  # Left-pointing Angle Brac..Right-pointing Angle Bra
        (0x023e9, 0x023ec,),  # Black Right-pointing Dou..Black Down-pointing Doub
        (0x023f0, 0x023f0,),  # Alarm Clock
        (0x023f3, 0x023f3,),  # Hourglass With Flowing Sand
        (0x025fd, 0x025fe,),  # White Medium Small Squar..Black Medium Small Squar
        (0x02614, 0x02615,),  # Umbrella With Rain Drops..Hot Beverage
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x02648, 0x02653,),  # Aries                   ..Pisces
        (0x0267f, 0x0267f,),  # Wheelchair Symbol
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x02693, 0x02693,),  # Anchor
        (0x026a1, 0x026a1,),  # High Voltage Sign
        (0x026aa, 0x026ab,),  # Medium White Circle     ..Medium Black Circle
        (0x026bd, 0x026be,),  # Soccer Ball             ..Baseball
        (0x026c4, 0x026c5,),  # Snowman Without Snow    ..Sun Behind Cloud
        (0x026ce, 0x026ce,),  # Ophiuchus
        (0x026d4, 0x026d4,),  # No Entry
        (0x026ea, 0x026ea,),  # Church
        (0x026f2, 0x026f3,),  # Fountain                ..Flag In Hole
        (0x026f5, 0x026f5,),  # Sailboat
        (0x026fa, 0x026fa,),  # Tent
        (0x026fd, 0x026fd,),  # Fuel Pump
        (0x02705, 0x02705,),  # White Heavy Check Mark
        (0x0270a, 0x0270b,),  # Raised Fist             ..Raised Hand
        (0x02728, 0x02728,),  # Sparkles
        (0x0274c, 0x0274c,),  # Cross Mark
        (0x0274e, 0x0274e,),  # Negative Squared Cross Mark
        (0x02753, 0x02755,),  # Black Question Mark Orna..White Exclamation Mark O
        (0x02757, 0x02757,),  # Heavy Exclamation Mark Symbol
        (0x02795, 0x02797,),  # Heavy Plus Sign         ..Heavy Division Sign
        (0x027b0, 0x027b0,),  # Curly Loop
        (0x027bf, 0x027bf,),  # Double Curly Loop
        (0x02b1b, 0x02b1c,),  # Black Large Square      ..White Large Square
        (0x02b50, 0x02b50,),  # White Medium Star
        (0x02b55, 0x02b55,),  # Heavy Large Circle
        (0x02e80, 0x02e99,),  # Cjk Radical Repeat      ..Cjk Radical Rap
        (0x02e9b, 0x02ef3,),  # Cjk Radical Choke       ..Cjk Radical C-simplified
        (0x02f00, 0x02fd5,),  # Kangxi Radical One      ..Kangxi Radical Flute
        (0x02ff0, 0x03029,),  # Ideographic Description ..Hangzhou Numeral Nine
        (0x03030, 0x0303e,),  # Wavy Dash               ..Ideographic Variation In
        (0x03041, 0x03096,),  # Hiragana Letter Small A ..Hiragana Letter Small Ke
        (0x0309b, 0x030ff,),  # Katakana-hiragana Voiced..Katakana Digraph Koto
        (0x03105, 0x0312f,),  # Bopomofo Letter B       ..Bopomofo Letter Nn
        (0x03131, 0x03163,),  # Hangul Letter Kiyeok    ..Hangul Letter I
        (0x03165, 0x0318e,),  # Hangul Letter Ssangnieun..Hangul Letter Araeae
        (0x03190, 0x031e5,),  # Ideographic Annotation L..Cjk Stroke Szp
        (0x031ef, 0x0321e,),  # Ideographic Description ..Parenthesized Korean Cha
        (0x03220, 0x03247,),  # Parenthesized Ideograph ..Circled Ideograph Koto
        (0x03250, 0x0a48c,),  # Partnership Sign        ..Yi Syllable Yyr
        (0x0a490, 0x0a4c6,),  # Yi Radical Qot          ..Yi Radical Ke
        (0x0a960, 0x0a97c,),  # Hangul Choseong Tikeut-m..Hangul Choseong Ssangyeo
        (0x0ac00, 0x0d7a3,),  # Hangul Syllable Ga      ..Hangul Syllable Hih
        (0x0f900, 0x0faff,),  # Cjk Compatibility Ideogr..(nil)
        (0x0fe10, 0x0fe19,),  # Presentation Form For Ve..Presentation Form For Ve
        (0x0fe30, 0x0fe52,),  # Presentation Form For Ve..Small Full Stop
        (0x0fe54, 0x0fe66,),  # Small Semicolon         ..Small Equals Sign
        (0x0fe68, 0x0fe6b,),  # Small Reverse Solidus   ..Small Commercial At
        (0x0ff01, 0x0ff60,),  # Fullwidth Exclamation Ma..Fullwidth Right White Pa
        (0x0ffe0, 0x0ffe6,),  # Fullwidth Cent Sign     ..Fullwidth Won Sign
        (0x16fe0, 0x16fe3,),  # Tangut Iteration Mark   ..Old Chinese Iteration Ma
        (0x16ff2, 0x16ff6,),  # Chinese Small Simplified..Yangqin Sign Slow Two Be
        (0x17000, 0x18cd5,),  # Tangut Ideograph-17000  ..Khitan Small Script Char
        (0x18cff, 0x18d1e,),  # Khitan Small Script Char..Tangut Ideograph-18d1e
        (0x18d80, 0x18df2,),  # Tangut Component-769    ..Tangut Component-883
        (0x1aff0, 0x1aff3,),  # Katakana Letter Minnan T..Katakana Letter Minnan T
        (0x1aff5, 0x1affb,),  # Katakana Letter Minnan T..Katakana Letter Minnan N
        (0x1affd, 0x1affe,),  # Katakana Letter Minnan N..Katakana Letter Minnan N
        (0x1b000, 0x1b122,),  # Katakana Letter Archaic ..Katakana Letter Archaic
        (0x1b132, 0x1b132,),  # Hiragana Letter Small Ko
        (0x1b150, 0x1b152,),  # Hiragana Letter Small Wi..Hiragana Letter Small Wo
        (0x1b155, 0x1b155,),  # Katakana Letter Small Ko
        (0x1b164, 0x1b167,),  # Katakana Letter Small Wi..Katakana Letter Small N
        (0x1b170, 0x1b2fb,),  # Nushu Character-1b170   ..Nushu Character-1b2fb
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f004, 0x1f004,),  # Mahjong Tile Red Dragon
        (0x1f0cf, 0x1f0cf,),  # Playing Card Black Joker
        (0x1f18e, 0x1f18e,),  # Negative Squared Ab
        (0x1f191, 0x1f19a,),  # Squared Cl              ..Squared Vs
        (0x1f1e6, 0x1f202,),  # Regional Indicator Symbo..Squared Katakana Sa
        (0x1f210, 0x1f23b,),  # Squared Cjk Unified Ideo..Squared Cjk Unified Ideo
        (0x1f240, 0x1f248,),  # Tortoise Shell Bracketed..Tortoise Shell Bracketed
        (0x1f250, 0x1f251,),  # Circled Ideograph Advant..Circled Ideograph Accept
        (0x1f260, 0x1f265,),  # Rounded Symbol For Fu   ..Rounded Symbol For Cai
        (0x1f300, 0x1f320,),  # Cyclone                 ..Shooting Star
        (0x1f32d, 0x1f335,),  # Hot Dog                 ..Cactus
        (0x1f337, 0x1f37c,),  # Tulip                   ..Baby Bottle
        (0x1f37e, 0x1f393,),  # Bottle With Popping Cork..Graduation Cap
        (0x1f3a0, 0x1f3ca,),  # Carousel Horse          ..Swimmer
        (0x1f3cf, 0x1f3d3,),  # Cricket Bat And Ball    ..Table Tennis Paddle And
        (0x1f3e0, 0x1f3f0,),  # House Building          ..European Castle
        (0x1f3f4, 0x1f3f4,),  # Waving Black Flag
        (0x1f3f8, 0x1f43e,),  # Badminton Racquet And Sh..Paw Prints
        (0x1f440, 0x1f440,),  # Eyes
        (0x1f442, 0x1f4fc,),  # Ear                     ..Videocassette
        (0x1f4ff, 0x1f53d,),  # Prayer Beads            ..Down-pointing Small Red
        (0x1f54b, 0x1f54e,),  # Kaaba                   ..Menorah With Nine Branch
        (0x1f550, 0x1f567,),  # Clock Face One Oclock   ..Clock Face Twelve-thirty
        (0x1f57a, 0x1f57a,),  # Man Dancing
        (0x1f595, 0x1f596,),  # Reversed Hand With Middl..Raised Hand With Part Be
        (0x1f5a4, 0x1f5a4,),  # Black Heart
        (0x1f5fb, 0x1f64f,),  # Mount Fuji              ..Person With Folded Hands
        (0x1f680, 0x1f6c5,),  # Rocket                  ..Left Luggage
        (0x1f6cc, 0x1f6cc,),  # Sleeping Accommodation
        (0x1f6d0, 0x1f6d2,),  # Place Of Worship        ..Shopping Trolley
        (0x1f6d5, 0x1f6d8,),  # Hindu Temple            ..Landslide
        (0x1f6dc, 0x1f6df,),  # Wireless                ..Ring Buoy
        (0x1f6eb, 0x1f6ec,),  # Airplane Departure      ..Airplane Arriving
        (0x1f6f4, 0x1f6fc,),  # Scooter                 ..Roller Skate
        (0x1f7e0, 0x1f7eb,),  # Large Orange Circle     ..Large Brown Square
        (0x1f7f0, 0x1f7f0,),  # Heavy Equals Sign
        (0x1f90c, 0x1f93a,),  # Pinched Fingers         ..Fencer
        (0x1f93c, 0x1f945,),  # Wrestlers               ..Goal Net
        (0x1f947, 0x1f9ff,),  # First Place Medal       ..Nazar Amulet
        (0x1fa70, 0x1fa7c,),  # Ballet Shoes            ..Crutch
        (0x1fa80, 0x1fa8a,),  # Yo-yo                   ..Trombone
        (0x1fa8e, 0x1fac6,),  # Treasure Chest          ..Fingerprint
        (0x1fac8, 0x1fac8,),  # Hairy Creature
        (0x1facd, 0x1fadc,),  # Orca                    ..Root Vegetable
        (0x1fadf, 0x1faea,),  # Splatter                ..Distorted Face
        (0x1faef, 0x1faf8,),  # Fight Cloud             ..Rightwards Pushing Hand
        (0x20000, 0x2fffd,),  # Cjk Unified Ideograph-20..(nil)
        (0x30000, 0x3fffd,),  # Cjk Unified Ideograph-30..(nil)
    ),
}



"""
Canonical terminal names and TERM/TERM_PROGRAM aliases.

This code generated by python wcwidth using ucs-detect project data.

Terminal multiplexers (tmux, zellij, libvterm, screen) are excluded because their displayed
presentation depends on the host terminal; cursor-position reports from ucs-detect testing are not
reliable indicators of actual width.
"""
# pylint: skip-file

KNOWN_TERMINALS = frozenset({
    'alacritty',
    'apple_terminal',
    'bobcat',
    'contour',
    'extraterm',
    'foot',
    'ghostty',
    'iterm2',
    'kitty',
    'konsole',
    'mintty',
    'mlterm',
    'pterm',
    'rio',
    'st',
    'terminology',
    'urxvt',
    'vte',
    'warp',
    'wezterm',
    'xterm',
    'xterm.js',
})

ALIASES = {
    'hyper': 'xterm.js',
    'iterm.app': 'iterm2',
    'putty': 'pterm',
    'rxvt': 'urxvt',
    'rxvt-unicode-256color': 'urxvt',
    'st-256color': 'st',
    'tabby': 'xterm.js',
    'vscode': 'xterm.js',
    'warpterminal': 'warp',
    'xterm-ghostty': 'ghostty',
    'xterm-kitty': 'kitty',
}



"""
Python 'wcwidth' module.

https://github.com/jquast/wcwidth
"""

# re-export common and outermost functions & definitions, even a few private
# ones, some for convenience, others for legacy, only the items in __all__ are
# documented as public API

__lazy_modules__ = [
    "wcwidth._clip",
    "wcwidth._wcswidth",
    "wcwidth._wcwidth",
    "wcwidth._width",
    "wcwidth.align",
    "wcwidth.bisearch",
    "wcwidth.escape_sequences",
    "wcwidth.grapheme",
    "wcwidth.hyperlink",
    "wcwidth.sgr_state",
    "wcwidth.table_ambiguous",
    "wcwidth.table_vs16",
    "wcwidth.table_wide",
    "wcwidth.table_zero",
    "wcwidth.text_sizing",
    "wcwidth.textwrap",
    "wcwidth.unicode_versions",
]

# local
from ._clip import clip
from .align import ljust, rjust, center
from ._width import width
from .bisearch import bisearch as _bisearch
from .grapheme import iter_graphemes, iter_graphemes_reverse, grapheme_boundary_before
from .textwrap import SequenceTextWrapper, wrap
from ._wcswidth import wcswidth, wcstwidth
from .hyperlink import Hyperlink, HyperlinkParams
from .sgr_state import propagate_sgr
from ._constants import list_term_programs
from .table_vs16 import VS16_NARROW_TO_WIDE
from .table_wide import WIDE_EASTASIAN
from .table_zero import ZERO_WIDTH
from .text_sizing import TextSizing, TextSizingParams
from .table_ambiguous import AMBIGUOUS_EASTASIAN
from .escape_sequences import iter_sequences, strip_sequences
from .unicode_versions import list_versions

# NOTE: this sort order is important for legacy import API compatibility before release 0.7.0
#
# On Python < 3.15 the legacy submodule is eagerly pre-imported for backward compatibility
# (populates sys.modules['wcwidth.wcwidth']).  On 3.15+ __lazy_modules__ handles all submodules; the
# legacy shim loads on-demand via file discovery when ``from wcwidth.wcwidth import ...`` is used.
if __import__('sys').version_info < (3, 15):
    # Pre-import the legacy submodule so that sys.modules['wcwidth.wcwidth'] is populated during
    # package initialization.  Without this, a later downstream dependent ``import wcwidth.wcwidth``
    # triggers on-disk file discovery which rebinds wcwidth.wcwidth from the function to the module
    # object.
    #
    # this is just a lot of carefulness for the original release that contained all functions in a
    # single 'wcwidth.py' file. Even though we always exposed our API at the top-level the preferred
    # 'from wcwidth import wcswidth', it was always possible to import them more directly,
    # 'from wcwidth.wcwidth import wcswidth'
    # -- and we make a lot of effort to allow any such import statements to continue to function.
    from . import wcwidth as _wcwidth_module  # isort:skip
from ._wcwidth import wcwidth, _wcmatch_version, _wcversion_value  # isort:skip  # pylint: disable=wrong-import-position


# The __all__ attribute defines the items exported from statement,
# 'from wcwidth import *', but also to say, "This is the public API".
__all__ = ('wcwidth', 'wcswidth', 'wcstwidth', 'width', 'iter_sequences', 'iter_graphemes',
           'iter_graphemes_reverse', 'grapheme_boundary_before',
           'ljust', 'rjust', 'center', 'wrap', 'clip', 'strip_sequences',
           'list_versions', 'list_term_programs', 'propagate_sgr',
           'Hyperlink', 'HyperlinkParams', 'TextSizing', 'TextSizingParams')

# Using 'hatchling', it does not seem to provide the pyproject.toml nicety, "dynamic = ['version']"
# like flit_core, maybe there is some better way but for now we have to duplicate it in both places
# Prefer the installed distribution version when available (helps test environments)
__version__ = '0.8.2'  # don't forget to also update pyproject.toml:version



"""
Sequence-aware text wrapping functions.

This module provides functions for wrapping text that may contain terminal escape sequences, with
proper handling of Unicode grapheme clusters and character display widths.
"""

from __future__ import annotations

# std imports
import secrets
import textwrap

from typing import TYPE_CHECKING, Optional

# local
from ._width import width as wcwidth_width
from .grapheme import iter_graphemes
from .hyperlink import HyperlinkParams
from .sgr_state import propagate_sgr as _propagate_sgr
from .escape_sequences import ZERO_WIDTH_PATTERN, iter_sequences

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Literal


class SequenceTextWrapper(textwrap.TextWrapper):
    """
    Sequence-aware text wrapper extending :class:`textwrap.TextWrapper`.

    This wrapper properly handles terminal escape sequences and Unicode grapheme clusters when
    calculating text width for wrapping.

    This implementation is based on the SequenceTextWrapper from the 'blessed' library, with
    contributions from Avram Lubkin and grayjk.

    The key difference from the blessed implementation is the addition of grapheme cluster support
    via :func:`~.iter_graphemes`, providing width calculation for ZWJ emoji sequences, VS-16 emojis
    and variations, regional indicator flags, and combining characters.

    OSC 8 hyperlinks are handled specially: when a hyperlink must span multiple lines, each line
    receives complete open/close sequences with a shared ``id`` parameter, ensuring terminals
    treat the fragments as a single hyperlink for hover underlining. If the original hyperlink
    already has an ``id`` parameter, it is preserved; otherwise, one is generated.
    """

    def __init__(self, width: int = 70, *,
                 control_codes: Literal['parse', 'strict', 'ignore'] = 'parse',
                 tabsize: int = 8,
                 ambiguous_width: int = 1,
                 term_program: bool | str = False,
                 **kwargs: Any) -> None:
        """
        Initialize the wrapper.

        :param width: Maximum line width in display cells.
        :param control_codes: How to handle control sequences (see :func:`~.width`).
        :param tabsize: Tab stop width for tab expansion.
        :param ambiguous_width: Width to use for East Asian Ambiguous (A) characters.
        :param term_program: Terminal software identifier for table correction.
            ``False`` (default) disables override lookup.  ``True`` reads the
            ``TERM_PROGRAM`` or ``TERM`` environment variable for auto-detection.
            Accepts a canonical terminal name matching :func:`list_term_programs`,
            such as from XTVERSION_, ENQ_, or ``TERM_PROGRAM``.

            .. versionadded:: 0.8.0
        :param kwargs: Additional arguments passed to :class:`textwrap.TextWrapper`.
        """
        super().__init__(width=width, **kwargs)
        self.control_codes = control_codes
        self.tabsize = tabsize
        self.ambiguous_width = ambiguous_width
        self.term_program = term_program

    @staticmethod
    def _next_hyperlink_id() -> str:
        """Generate unique hyperlink id as 8-character hex string."""
        return secrets.token_hex(4)

    def _width(self, text: str) -> int:
        """Measure text width accounting for sequences."""
        return wcwidth_width(text, control_codes=self.control_codes, tabsize=self.tabsize,
                             ambiguous_width=self.ambiguous_width,
                             term_program=self.term_program)

    def _strip_sequences(self, text: str) -> str:
        """Strip all terminal sequences from text."""
        result = []
        for segment, is_seq in iter_sequences(text):
            if not is_seq:
                result.append(segment)
        return ''.join(result)

    def _extract_sequences(self, text: str) -> str:
        """Extract only terminal sequences from text."""
        result = []
        for segment, is_seq in iter_sequences(text):
            if is_seq:
                result.append(segment)
        return ''.join(result)

    def _split(self, text: str) -> list[str]:  # pylint: disable=too-many-locals
        r"""
        Sequence-aware variant of :meth:`textwrap.TextWrapper._split`.

        This method ensures that terminal escape sequences don't interfere with the text splitting
        logic, particularly for hyphen-based word breaking. It builds a position mapping from
        stripped text to original text, calls the parent's _split on stripped text, then maps chunks
        back.

        OSC hyperlink sequences are treated as word boundaries::

            >>> wrap('foo \x1b]8;;https://example.com\x07link\x1b]8;;\x07 bar', 6)
            ['foo', '\x1b]8;;https://example.com\x07link\x1b]8;;\x07', 'bar']

        Both BEL (``\x07``) and ST (``\x1b\\``) terminators are supported.
        """
        # pylint: disable=too-many-locals,too-many-branches
        # Build a mapping from stripped text positions to original text positions.
        #
        # Track where each character ENDS so that sequences between characters
        # attach to the following text (not preceding text). This ensures sequences
        # aren't lost when whitespace is dropped.
        #
        # char_end[i] = position in original text right after the i-th stripped char
        char_end: list[int] = []
        stripped_text = ''
        original_pos = 0
        prev_was_hyperlink_close = False

        for segment, is_seq in iter_sequences(text):
            if not is_seq:
                # Conditionally insert space after hyperlink close to force word boundary
                if prev_was_hyperlink_close and segment and not segment[0].isspace():
                    stripped_text += ' '
                    char_end.append(original_pos)
                for char in segment:
                    original_pos += 1
                    char_end.append(original_pos)
                    stripped_text += char
                prev_was_hyperlink_close = False
            else:
                is_hyperlink_close = segment.startswith(('\x1b]8;;\x1b\\', '\x1b]8;;\x07'))

                # Conditionally insert space before OSC sequences to artificially create word
                # boundary, but *not* before hyperlink close sequences, to ensure hyperlink is
                # terminated on the same line.
                if (segment.startswith('\x1b]') and stripped_text and not
                        stripped_text[-1].isspace()):
                    if not is_hyperlink_close:
                        stripped_text += ' '
                        char_end.append(original_pos)

                # Escape sequences advance position but don't add to stripped text
                original_pos += len(segment)
                prev_was_hyperlink_close = is_hyperlink_close

        # Add sentinel for final position
        char_end.append(original_pos)

        # Use parent's _split on the stripped text
        # pylint: disable-next=protected-access
        stripped_chunks = textwrap.TextWrapper._split(self, stripped_text)

        # Handle text that contains only sequences (no visible characters).
        # Return the sequences as a single chunk to preserve them.
        if not stripped_chunks and text:
            return [text]

        # Map the chunks back to the original text with sequences
        result: list[str] = []
        stripped_pos = 0
        num_chunks = len(stripped_chunks)

        for idx, chunk in enumerate(stripped_chunks):
            chunk_len = len(chunk)

            # Start is where previous character ended (or 0 for first chunk)
            start_orig = 0 if stripped_pos == 0 else char_end[stripped_pos - 1]

            # End is where next character starts. For last chunk, use sentinel
            # to include any trailing sequences.
            if idx == num_chunks - 1:
                end_orig = char_end[-1]  # sentinel includes trailing sequences
            else:
                end_orig = char_end[stripped_pos + chunk_len - 1]

            # Extract the corresponding portion from the original text
            # Skip empty chunks (from virtual spaces inserted at OSC boundaries)
            if start_orig != end_orig:
                result.append(text[start_orig:end_orig])
            stripped_pos += chunk_len

        return result

    def _wrap_chunks(self, chunks: list[str]) -> list[str]:  # pylint: disable=too-many-branches
        """
        Wrap chunks into lines using sequence-aware width.

        Override TextWrapper._wrap_chunks to use _width instead of len. Follows stdlib's algorithm:
        greedily fill lines, handle long words.  Also handle OSC hyperlink processing. When
        hyperlinks span multiple lines, each line gets complete open/close sequences with matching
        id parameters for hover underlining continuity per OSC 8 spec.
        """
        # pylint: disable=too-many-branches,too-many-statements,too-complex,too-many-locals
        # pylint: disable=too-many-nested-blocks
        # the hyperlink code in particular really pushes the complexity rating of this method.
        # preferring to keep it "all in one method" because of so much local state and manipulation.
        if not chunks:
            return []

        if self.max_lines is not None:
            if self.max_lines > 1:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent
            if (self._width(indent)
                    + self._width(self.placeholder.lstrip())
                    > self.width):
                raise ValueError("placeholder too large for max width")

        lines: list[str] = []
        is_first_line = True

        hyperlink_state: Optional[HyperlinkParams] = None
        # Track the id we're using for the current hyperlink continuation
        current_hyperlink_id: Optional[str] = None

        # Arrange in reverse order so items can be efficiently popped
        chunks = list(reversed(chunks))

        while chunks:
            current_line: list[str] = []
            current_width = 0

            # Get the indent and available width for current line
            indent = self.initial_indent if is_first_line else self.subsequent_indent
            line_width = self.width - self._width(indent)

            # If continuing a hyperlink from previous line, prepend open sequence
            if hyperlink_state is not None:
                open_seq = HyperlinkParams(
                    url=hyperlink_state.url,
                    params=hyperlink_state.params,
                    terminator=hyperlink_state.terminator,
                ).make_open()
                chunks[-1] = open_seq + chunks[-1]

            # Drop leading whitespace (except at very start)
            # When dropping, transfer any sequences to the next chunk.
            # Only drop if there's actual whitespace text, not if it's only sequences.
            stripped = self._strip_sequences(chunks[-1])
            if self.drop_whitespace and lines and stripped and not stripped.strip():
                sequences = self._extract_sequences(chunks[-1])
                del chunks[-1]
                if sequences and chunks:
                    chunks[-1] = sequences + chunks[-1]

            # Greedily add chunks that fit
            while chunks:
                chunk = chunks[-1]
                chunk_width = self._width(chunk)

                if current_width + chunk_width <= line_width:
                    current_line.append(chunks.pop())
                    current_width += chunk_width
                else:
                    break

            # Handle chunk that's too long for any line
            if chunks and self._width(chunks[-1]) > line_width:
                self._handle_long_word(
                    chunks, current_line, current_width, line_width
                )
                current_width = self._width(''.join(current_line))
                # Remove any empty chunks left by _handle_long_word
                while chunks and not chunks[-1]:
                    del chunks[-1]

            # Drop trailing whitespace
            # When dropping, transfer any sequences to the previous chunk.
            # Only drop if there's actual whitespace text, not if it's only sequences.
            stripped_last = self._strip_sequences(current_line[-1]) if current_line else ''
            if (self.drop_whitespace and current_line and
                    stripped_last and not stripped_last.strip()):
                sequences = self._extract_sequences(current_line[-1])
                current_width -= self._width(current_line[-1])
                del current_line[-1]
                if sequences and current_line:
                    current_line[-1] = current_line[-1] + sequences

            if current_line:
                # Check whether this is a normal append or max_lines
                # truncation. Matches stdlib textwrap precedence:
                # normal if max_lines not set, not yet reached, or no
                # remaining visible content that would need truncation.
                no_more_content = (
                    not chunks or
                    self.drop_whitespace and
                    len(chunks) == 1 and
                    not self._strip_sequences(chunks[0]).strip()
                )
                if (self.max_lines is None or
                        len(lines) + 1 < self.max_lines or
                        no_more_content
                        and current_width <= line_width):
                    line_content = ''.join(current_line)

                    # Track hyperlink state through this line's content
                    new_state = self._track_hyperlink_state(line_content, hyperlink_state)

                    # If we end inside a hyperlink, append close sequence
                    if new_state is not None:
                        # Ensure we have an id for continuation
                        if current_hyperlink_id is None:
                            if 'id=' in new_state.params:
                                current_hyperlink_id = new_state.params
                            elif new_state.params:
                                # Prepend id to existing params. Per OSC 8 spec, params can have
                                # multiple key=value pairs separated by ':'.
                                current_hyperlink_id = (
                                    f'id={self._next_hyperlink_id()}:{new_state.params}')
                            else:
                                current_hyperlink_id = f'id={self._next_hyperlink_id()}'
                        line_content += HyperlinkParams(
                            terminator=new_state.terminator, url='').make_close()

                        # Also need to inject the id into the opening
                        # sequence if it didn't have one
                        if 'id=' not in new_state.params:
                            # Find and replace the original open sequence with one that has id
                            old_open = HyperlinkParams(
                                url=new_state.url,
                                params=new_state.params,
                                terminator=new_state.terminator,
                            ).make_open()
                            new_open = HyperlinkParams(
                                url=new_state.url,
                                params=current_hyperlink_id,
                                terminator=new_state.terminator,
                            ).make_open()
                            line_content = line_content.replace(old_open, new_open, 1)

                        # Update state for next line, using computed id
                        hyperlink_state = HyperlinkParams(
                            new_state.url, current_hyperlink_id, new_state.terminator)
                    else:
                        hyperlink_state = None
                        current_hyperlink_id = None  # Reset id when hyperlink closes

                    # Strip trailing whitespace when drop_whitespace is enabled
                    # (matches CPython #140627 fix behavior)
                    if self.drop_whitespace:
                        line_content = line_content.rstrip()
                    lines.append(indent + line_content)
                    is_first_line = False
                else:
                    # max_lines reached with remaining content.
                    # pop chunks until placeholder fits, then break.
                    placeholder_w = self._width(self.placeholder)
                    while current_line:
                        last_text = self._strip_sequences(current_line[-1])
                        if (last_text.strip()
                                and current_width + placeholder_w <= line_width):
                            line_content = ''.join(current_line)
                            new_state = self._track_hyperlink_state(
                                line_content, hyperlink_state)
                            if new_state is not None:
                                line_content += HyperlinkParams(
                                    terminator=new_state.terminator, url='').make_close()
                            lines.append(indent + line_content + self.placeholder)
                            break
                        current_width -= self._width(current_line[-1])
                        del current_line[-1]
                    else:
                        if lines:
                            prev_line = self._rstrip_visible(lines[-1])
                            if (self._width(prev_line) + placeholder_w
                                    <= self.width):
                                lines[-1] = prev_line + self.placeholder
                                break
                        lines.append(indent + self.placeholder.lstrip())
                    break

        return lines

    def _track_hyperlink_state(
            self, text: str,
            state: Optional[HyperlinkParams]) -> Optional[HyperlinkParams]:
        """
        Track hyperlink state through text.

        :param text: Text to scan for hyperlink sequences.
        :param state: Current state or None if outside hyperlink.
        :returns: Updated state after processing text.
        """
        for segment, is_seq in iter_sequences(text):
            if is_seq:
                parsed_link = HyperlinkParams.parse(segment)
                if parsed_link is not None and parsed_link.url:  # has URL = open
                    state = parsed_link
                elif segment.startswith(('\x1b]8;;\x1b\\', '\x1b]8;;\x07')):  # close
                    state = None
        return state

    def _handle_long_word(self, reversed_chunks: list[str],
                          cur_line: list[str], cur_len: int,
                          width: int) -> None:
        """
        Sequence-aware :meth:`textwrap.TextWrapper._handle_long_word`.

        This method ensures that word boundaries are not broken mid-sequence, and respects grapheme
        cluster boundaries when breaking long words.
        """
        if width < 1:
            space_left = 1
        else:
            space_left = width - cur_len

        chunk = reversed_chunks[-1]

        if self.break_long_words:
            break_at_hyphen = False
            hyphen_end = 0

            # Handle break_on_hyphens: find last hyphen within space_left
            if self.break_on_hyphens:
                # Strip sequences to find hyphen in logical text
                stripped = self._strip_sequences(chunk)
                if len(stripped) > space_left:
                    # Find last hyphen in the portion that fits
                    hyphen_pos = stripped.rfind('-', 0, space_left)
                    if hyphen_pos > 0 and any(c != '-' for c in stripped[:hyphen_pos]):
                        # Map back to original position including sequences
                        hyphen_end = self._map_stripped_pos_to_original(chunk, hyphen_pos + 1)
                        break_at_hyphen = True

            # Break at grapheme boundaries to avoid splitting multi-codepoint characters
            if break_at_hyphen:
                actual_end = hyphen_end
            else:
                actual_end = self._find_break_position(chunk, space_left)
                # If no progress possible (e.g., wide char exceeds line width),
                # force at least one grapheme to avoid infinite loop.
                # Only force when cur_line is empty; if line has content,
                # appending nothing is safe and the line will be committed.
                if actual_end == 0 and not cur_line:
                    actual_end = self._find_first_grapheme_end(chunk)
            cur_line.append(chunk[:actual_end])
            reversed_chunks[-1] = chunk[actual_end:]

        elif not cur_line:
            cur_line.append(reversed_chunks.pop())

    def _map_stripped_pos_to_original(self, text: str, stripped_pos: int) -> int:
        """Map a position in stripped text back to original text position."""
        stripped_idx = 0
        original_idx = 0

        for segment, is_seq in iter_sequences(text):
            if is_seq:
                original_idx += len(segment)
            elif stripped_idx + len(segment) > stripped_pos:
                # Position is within this segment
                return original_idx + (stripped_pos - stripped_idx)
            else:
                stripped_idx += len(segment)
                original_idx += len(segment)

        # Caller guarantees stripped_pos < total stripped chars, so we always
        # return from within the loop. This line satisfies the type checker.
        return original_idx  # pragma: no cover

    def _find_break_position(self, text: str, max_width: int) -> int:
        """Find string index in text that fits within max_width cells."""
        idx = 0
        width_so_far = 0

        while idx < len(text):
            char = text[idx]

            # Skip escape sequences (they don't add width)
            if char == '\x1b':
                match = ZERO_WIDTH_PATTERN.match(text, idx)
                if match:
                    idx = match.end()
                    continue

            # Get grapheme (use start= to avoid slice allocation)
            grapheme = next(iter_graphemes(text, start=idx))

            grapheme_width = self._width(grapheme)
            if width_so_far + grapheme_width > max_width:
                return idx  # Found break point

            width_so_far += grapheme_width
            idx += len(grapheme)

        # Caller guarantees chunk_width > max_width, so a grapheme always
        # exceeds and we return from within the loop. Type checker requires this.
        return idx  # pragma: no cover

    def _find_first_grapheme_end(self, text: str) -> int:
        """Find the end position of the first grapheme."""
        return len(next(iter_graphemes(text)))

    def _rstrip_visible(self, text: str) -> str:
        """Strip trailing visible whitespace, preserving trailing sequences."""
        segments = list(iter_sequences(text))
        last_vis = -1
        for i, (segment, is_seq) in enumerate(segments):
            if not is_seq and segment.rstrip():
                last_vis = i
        if last_vis == -1:
            return ''
        result = []
        for i, (segment, is_seq) in enumerate(segments):
            if i < last_vis:
                result.append(segment)
            elif i == last_vis:
                result.append(segment.rstrip())
            elif is_seq:
                result.append(segment)
        return ''.join(result)


def wrap(text: str, width: int = 70, *,
         control_codes: Literal['parse', 'strict', 'ignore'] = 'parse',
         tabsize: int = 8,
         expand_tabs: bool = True,
         replace_whitespace: bool = True,
         ambiguous_width: int = 1,
         term_program: bool | str = False,
         initial_indent: str = '',
         subsequent_indent: str = '',
         fix_sentence_endings: bool = False,
         break_long_words: bool = True,
         break_on_hyphens: bool = True,
         drop_whitespace: bool = True,
         max_lines: Optional[int] = None,
         placeholder: str = ' [...]',
         propagate_sgr: bool = True) -> list[str]:
    r"""
    Wrap text to fit within given width, returning a list of wrapped lines.

    Like :func:`textwrap.wrap`, but measures width in display cells rather than
    characters, correctly handling wide characters, combining marks, and terminal
    escape sequences.

    :param text: Text to wrap, may contain terminal sequences.
    :param width: Maximum line width in display cells.
    :param control_codes: How to handle terminal sequences (see :func:`~.width`).
    :param tabsize: Tab stop width for tab expansion.
    :param expand_tabs: If True (default), tab characters are expanded
        to spaces using ``tabsize``.
    :param replace_whitespace: If True (default), each whitespace character
        is replaced with a single space after tab expansion. When False,
        control whitespace like ``\n`` has zero display width (unlike
        :func:`textwrap.wrap` which counts ``len()``), so wrap points
        may differ from stdlib for non-space whitespace characters.
    :param ambiguous_width: Width to use for East Asian Ambiguous (A)
        characters. Default is ``1`` (narrow). Set to ``2`` for CJK contexts.
    :param term_program: Terminal software identifier for table correction.
        ``False`` (default) disables override lookup.  ``True`` reads the
        ``TERM_PROGRAM`` or ``TERM`` environment variable for auto-detection.
        Accepts a canonical terminal name matching :func:`list_term_programs`,
        such as from XTVERSION_, ENQ_, or ``TERM_PROGRAM``.

        .. versionadded:: 0.8.0
    :param initial_indent: String prepended to first line.
    :param subsequent_indent: String prepended to subsequent lines.
    :param fix_sentence_endings: If True, ensure sentences are always
        separated by exactly two spaces.
    :param break_long_words: If True, break words longer than width.
    :param break_on_hyphens: If True, allow breaking at hyphens.
    :param drop_whitespace: If True (default), whitespace at the beginning
        and end of each line (after wrapping but before indenting) is dropped.
        Set to False to preserve whitespace.
    :param max_lines: If set, output contains at most this many lines, with
        ``placeholder`` appended to the last line if the text was truncated.
    :param placeholder: String appended to the last line when text is
        truncated by ``max_lines``. Default is ``' [...]'``.
    :param propagate_sgr: If True (default), SGR (terminal styling) sequences
        are propagated across wrapped lines. Each line ends with a reset
        sequence and the next line begins with the active style restored.
    :returns: List of wrapped lines without trailing newlines.

    SGR (terminal styling) sequences are propagated across wrapped lines
    by default. Each line ends with a reset sequence and the next line
    begins with the active style restored::

        >>> wrap('\x1b[1;34mHello world\x1b[0m', width=6)
        ['\x1b[1;34mHello\x1b[0m', '\x1b[1;34mworld\x1b[0m']

    Set ``propagate_sgr=False`` to disable this behavior.

    Like :func:`textwrap.wrap`, newlines in the input text are treated as
    whitespace and collapsed. To preserve paragraph breaks, wrap each
    paragraph separately::

        >>> text = 'First line.\nSecond line.'
        >>> wrap(text, 40)  # newline collapsed to space
        ['First line. Second line.']
        >>> [line for para in text.split('\n')
        ...  for line in (wrap(para, 40) if para else [''])]
        ['First line.', 'Second line.']

    .. seealso::

       :func:`textwrap.wrap`, :class:`textwrap.TextWrapper`
           Standard library text wrapping (character-based).

       :class:`.SequenceTextWrapper`
           Class interface for advanced wrapping options.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.5.0
       Added ``propagate_sgr`` parameter (default True).

    .. versionchanged:: 0.6.0
       Added ``expand_tabs``, ``replace_whitespace``, ``fix_sentence_endings``,
       ``drop_whitespace``, ``max_lines``, and ``placeholder`` parameters.

    Example::

        >>> from wcwidth import wrap
        >>> wrap('hello world', 5)
        ['hello', 'world']
        >>> wrap('中文字符', 4)  # CJK characters (2 cells each)
        ['中文', '字符']
    """
    # pylint: disable=too-many-arguments,too-many-locals
    wrapper = SequenceTextWrapper(
        width=width,
        control_codes=control_codes,
        tabsize=tabsize,
        expand_tabs=expand_tabs,
        replace_whitespace=replace_whitespace,
        ambiguous_width=ambiguous_width,
        term_program=term_program,
        initial_indent=initial_indent,
        subsequent_indent=subsequent_indent,
        fix_sentence_endings=fix_sentence_endings,
        break_long_words=break_long_words,
        break_on_hyphens=break_on_hyphens,
        drop_whitespace=drop_whitespace,
        max_lines=max_lines,
        placeholder=placeholder,
    )
    lines = wrapper.wrap(text)

    if propagate_sgr:
        lines = _propagate_sgr(lines)

    return lines



"""
Exports function list_versions() for unicode version level support.

This code generated by python wcwidth project.
"""

from __future__ import annotations


def list_versions() -> tuple[str, ...]:
    """
    Return Unicode version levels supported by this module release.

    .. versionchanged:: 0.5.0
       Now returns a single-element tuple containing only the latest version.

    :returns: Supported Unicode version numbers in ascending sorted order.
    """
    return (
        "17.0.0",
    )



"""
Terminal override tables keyed by canonical terminal software name.

This code generated by python wcwidth using ucs-detect project data.
"""
# pylint: skip-file
from __future__ import annotations

# For terminals: rio
_SET_WIDE_OVERRIDES_243C8ED7: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1f6d8, 0x1f6d8,),  # Landslide
        (0x1fa8a, 0x1fa8a,),  # Trombone
        (0x1fa8e, 0x1fa8e,),  # Treasure Chest
        (0x1fac8, 0x1fac8,),  # Hairy Creature
        (0x1facd, 0x1facd,),  # Orca
        (0x1faea, 0x1faea,),  # Distorted Face
        (0x1faef, 0x1faef,),  # Fight Cloud
    ),
}

# For terminals: konsole
_SET_WIDE_OVERRIDES_25894E07: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x02ffc, 0x02fff,),  # Ideographic Description ..Ideographic Description
        (0x031e4, 0x031e5,),  # Cjk Stroke Hxg          ..Cjk Stroke Szp
        (0x031ef, 0x031ef,),  # Ideographic Description Character Subtraction
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f6d8, 0x1f6d8,),  # Landslide
        (0x1fa89, 0x1fa8a,),  # Harp                    ..Trombone
        (0x1fa8e, 0x1fa8f,),  # Treasure Chest          ..Shovel
        (0x1fabe, 0x1fabe,),  # Leafless Tree
        (0x1fac6, 0x1fac6,),  # Fingerprint
        (0x1fac8, 0x1fac8,),  # Hairy Creature
        (0x1facd, 0x1facd,),  # Orca
        (0x1fadc, 0x1fadc,),  # Root Vegetable
        (0x1fadf, 0x1fadf,),  # Splatter
        (0x1fae9, 0x1faea,),  # Face With Bags Under Eye..Distorted Face
        (0x1faef, 0x1faef,),  # Fight Cloud
    ),
}

# For terminals: extraterm
_SET_WIDE_OVERRIDES_6C1578E8: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x02ffc, 0x02fff,),  # Ideographic Description ..Ideographic Description
        (0x031e4, 0x031e5,),  # Cjk Stroke Hxg          ..Cjk Stroke Szp
        (0x031ef, 0x031ef,),  # Ideographic Description Character Subtraction
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1f6d8, 0x1f6d8,),  # Landslide
        (0x1f6dc, 0x1f6dc,),  # Wireless
        (0x1fa75, 0x1fa77,),  # Light Blue Heart        ..Pink Heart
        (0x1fa87, 0x1fa8a,),  # Maracas                 ..Trombone
        (0x1fa8e, 0x1fa8f,),  # Treasure Chest          ..Shovel
        (0x1faad, 0x1faaf,),  # Folding Hand Fan        ..Khanda
        (0x1fabb, 0x1fabf,),  # Hyacinth                ..Goose
        (0x1fac6, 0x1fac6,),  # Fingerprint
        (0x1fac8, 0x1fac8,),  # Hairy Creature
        (0x1facd, 0x1facf,),  # Orca                    ..Donkey
        (0x1fada, 0x1fadc,),  # Ginger Root             ..Root Vegetable
        (0x1fadf, 0x1fadf,),  # Splatter
        (0x1fae8, 0x1faea,),  # Shaking Face            ..Distorted Face
        (0x1faef, 0x1faef,),  # Fight Cloud
        (0x1faf7, 0x1faf8,),  # Leftwards Pushing Hand  ..Rightwards Pushing Hand
    ),
}

# For terminals: alacritty, apple_terminal, pterm, st, urxvt, xterm
_SET_WIDE_OVERRIDES_6E5EF6CC: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
    ),
}

# For terminals: kitty
_SET_WIDE_OVERRIDES_73E5B60D: dict[str, tuple[tuple[int, int], ...]] = {
    'zeroer': (
        (0x1f3fb, 0x1f3ff,),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
    ),
}

# For terminals: mintty
_SET_WIDE_OVERRIDES_79ED351B: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x02ffc, 0x02fff,),  # Ideographic Description ..Ideographic Description
        (0x031e4, 0x031e5,),  # Cjk Stroke Hxg          ..Cjk Stroke Szp
        (0x031ef, 0x031ef,),  # Ideographic Description Character Subtraction
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1fa89, 0x1fa89,),  # Harp
        (0x1fa8f, 0x1fa8f,),  # Shovel
        (0x1fabe, 0x1fabe,),  # Leafless Tree
        (0x1fac6, 0x1fac6,),  # Fingerprint
        (0x1fadc, 0x1fadc,),  # Root Vegetable
        (0x1fadf, 0x1fadf,),  # Splatter
        (0x1fae9, 0x1fae9,),  # Face With Bags Under Eyes
    ),
}

# For terminals: bobcat
_SET_WIDE_OVERRIDES_8A2DF0C6: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x03030, 0x03030,),  # Wavy Dash
        (0x0303d, 0x0303d,),  # Part Alternation Mark
        (0x03297, 0x03297,),  # Circled Ideograph Congratulation
        (0x03299, 0x03299,),  # Circled Ideograph Secret
        (0x1f202, 0x1f202,),  # Squared Katakana Sa
        (0x1f237, 0x1f237,),  # Squared Cjk Unified Ideograph-6708
    ),
    'zeroer': (
        (0x1f3fb, 0x1f3ff,),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
    ),
}

# For terminals: foot
_SET_WIDE_OVERRIDES_98684D9B: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1f3fb, 0x1f3ff,),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
    ),
}

# For terminals: xterm.js
_SET_WIDE_OVERRIDES_9D8AD458: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x02ffc, 0x02fff,),  # Ideographic Description ..Ideographic Description
        (0x031bb, 0x031bf,),  # Bopomofo Final Letter G ..Bopomofo Letter Ah
        (0x031e4, 0x031e5,),  # Cjk Stroke Hxg          ..Cjk Stroke Szp
        (0x031ef, 0x031ef,),  # Ideographic Description Character Subtraction
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x1aff0, 0x1aff3,),  # Katakana Letter Minnan T..Katakana Letter Minnan T
        (0x1aff5, 0x1affb,),  # Katakana Letter Minnan T..Katakana Letter Minnan N
        (0x1affd, 0x1affe,),  # Katakana Letter Minnan N..Katakana Letter Minnan N
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1f6d6, 0x1f6d8,),  # Hut                     ..Landslide
        (0x1f6dc, 0x1f6df,),  # Wireless                ..Ring Buoy
        (0x1f6fb, 0x1f6fc,),  # Pickup Truck            ..Roller Skate
        (0x1f7f0, 0x1f7f0,),  # Heavy Equals Sign
        (0x1f90c, 0x1f90c,),  # Pinched Fingers
        (0x1f972, 0x1f972,),  # Smiling Face With Tear
        (0x1f977, 0x1f979,),  # Ninja                   ..Face Holding Back Tears
        (0x1f9a3, 0x1f9a4,),  # Mammoth                 ..Dodo
        (0x1f9ab, 0x1f9ad,),  # Beaver                  ..Seal
        (0x1f9cb, 0x1f9cc,),  # Bubble Tea              ..Troll
        (0x1fa74, 0x1fa77,),  # Thong Sandal            ..Pink Heart
        (0x1fa7b, 0x1fa7c,),  # X-ray                   ..Crutch
        (0x1fa83, 0x1fa8a,),  # Boomerang               ..Trombone
        (0x1fa8e, 0x1fa8f,),  # Treasure Chest          ..Shovel
        (0x1fa96, 0x1fac6,),  # Military Helmet         ..Fingerprint
        (0x1fac8, 0x1fac8,),  # Hairy Creature
        (0x1facd, 0x1fadc,),  # Orca                    ..Root Vegetable
        (0x1fadf, 0x1faea,),  # Splatter                ..Distorted Face
        (0x1faef, 0x1faf8,),  # Fight Cloud             ..Rightwards Pushing Hand
    ),
}

# For terminals: vte, wezterm
_SET_WIDE_OVERRIDES_A5DF05D4: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x02ffc, 0x02fff,),  # Ideographic Description ..Ideographic Description
        (0x031e4, 0x031e5,),  # Cjk Stroke Hxg          ..Cjk Stroke Szp
        (0x031ef, 0x031ef,),  # Ideographic Description Character Subtraction
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1fa89, 0x1fa89,),  # Harp
        (0x1fa8f, 0x1fa8f,),  # Shovel
        (0x1fabe, 0x1fabe,),  # Leafless Tree
        (0x1fac6, 0x1fac6,),  # Fingerprint
        (0x1fadc, 0x1fadc,),  # Root Vegetable
        (0x1fadf, 0x1fadf,),  # Splatter
        (0x1fae9, 0x1fae9,),  # Face With Bags Under Eyes
    ),
}

# For terminals: terminology
_SET_WIDE_OVERRIDES_C908F105: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x0fada, 0x0faff,),  # (nil)
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1f6d8, 0x1f6d8,),  # Landslide
        (0x1f6dc, 0x1f6df,),  # Wireless                ..Ring Buoy
        (0x1f7f0, 0x1f7f0,),  # Heavy Equals Sign
        (0x1fad7, 0x1fadc,),  # Pouring Liquid          ..Root Vegetable
        (0x1fadf, 0x1faea,),  # Splatter                ..Distorted Face
        (0x1faef, 0x1faf8,),  # Fight Cloud             ..Rightwards Pushing Hand
    ),
}

# For terminals: ghostty, iterm2
_SET_WIDE_OVERRIDES_CF30CE25: dict[str, tuple[tuple[int, int], ...]] = {
    'zeroer': (
        (0x0115f, 0x0115f,),  # Hangul Choseong Filler
    ),
}

# For terminals: mlterm, warp
_SET_WIDE_OVERRIDES_D289A768: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x02630, 0x02637,),  # Trigram For Heaven      ..Trigram For Earth
        (0x0268a, 0x0268f,),  # Monogram For Yang       ..Digram For Greater Yin
        (0x031e4, 0x031e5,),  # Cjk Stroke Hxg          ..Cjk Stroke Szp
        (0x04dc0, 0x04dff,),  # Hexagram For The Creativ..Hexagram For Before Comp
        (0x1d300, 0x1d356,),  # Monogram For Earth      ..Tetragram For Fostering
        (0x1d360, 0x1d376,),  # Counting Rod Unit Digit ..Ideographic Tally Mark F
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
        (0x1f6d8, 0x1f6d8,),  # Landslide
        (0x1fa89, 0x1fa8a,),  # Harp                    ..Trombone
        (0x1fa8e, 0x1fa8f,),  # Treasure Chest          ..Shovel
        (0x1fabe, 0x1fabe,),  # Leafless Tree
        (0x1fac6, 0x1fac6,),  # Fingerprint
        (0x1fac8, 0x1fac8,),  # Hairy Creature
        (0x1facd, 0x1facd,),  # Orca
        (0x1fadc, 0x1fadc,),  # Root Vegetable
        (0x1fadf, 0x1fadf,),  # Splatter
        (0x1fae9, 0x1faea,),  # Face With Bags Under Eye..Distorted Face
        (0x1faef, 0x1faef,),  # Fight Cloud
    ),
}

WIDE_OVERRIDES: dict[str, dict[str, tuple[tuple[int, int], ...]]] = {
    'alacritty': _SET_WIDE_OVERRIDES_6E5EF6CC,
    'apple_terminal': _SET_WIDE_OVERRIDES_6E5EF6CC,
    'bobcat': _SET_WIDE_OVERRIDES_8A2DF0C6,
    'extraterm': _SET_WIDE_OVERRIDES_6C1578E8,
    'foot': _SET_WIDE_OVERRIDES_98684D9B,
    'ghostty': _SET_WIDE_OVERRIDES_CF30CE25,
    'iterm2': _SET_WIDE_OVERRIDES_CF30CE25,
    'kitty': _SET_WIDE_OVERRIDES_73E5B60D,
    'konsole': _SET_WIDE_OVERRIDES_25894E07,
    'mintty': _SET_WIDE_OVERRIDES_79ED351B,
    'mlterm': _SET_WIDE_OVERRIDES_D289A768,
    'pterm': _SET_WIDE_OVERRIDES_6E5EF6CC,
    'rio': _SET_WIDE_OVERRIDES_243C8ED7,
    'st': _SET_WIDE_OVERRIDES_6E5EF6CC,
    'terminology': _SET_WIDE_OVERRIDES_C908F105,
    'urxvt': _SET_WIDE_OVERRIDES_6E5EF6CC,
    'vte': _SET_WIDE_OVERRIDES_A5DF05D4,
    'warp': _SET_WIDE_OVERRIDES_D289A768,
    'wezterm': _SET_WIDE_OVERRIDES_A5DF05D4,
    'xterm': _SET_WIDE_OVERRIDES_6E5EF6CC,
    'xterm.js': _SET_WIDE_OVERRIDES_9D8AD458,
}

# For terminals: alacritty, apple_terminal, extraterm, foot, mlterm, pterm, rio, st, terminology, urxvt, vte, warp, wezterm, xterm, xterm.js
_SET_SRI_OVERRIDES_6E5EF6CC: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x1f1e6, 0x1f1ff,),  # Regional Indicator Symbo..Regional Indicator Symbo
    ),
}

SRI_OVERRIDES: dict[str, dict[str, tuple[tuple[int, int], ...]]] = {
    'alacritty': _SET_SRI_OVERRIDES_6E5EF6CC,
    'apple_terminal': _SET_SRI_OVERRIDES_6E5EF6CC,
    'extraterm': _SET_SRI_OVERRIDES_6E5EF6CC,
    'foot': _SET_SRI_OVERRIDES_6E5EF6CC,
    'mlterm': _SET_SRI_OVERRIDES_6E5EF6CC,
    'pterm': _SET_SRI_OVERRIDES_6E5EF6CC,
    'rio': _SET_SRI_OVERRIDES_6E5EF6CC,
    'st': _SET_SRI_OVERRIDES_6E5EF6CC,
    'terminology': _SET_SRI_OVERRIDES_6E5EF6CC,
    'urxvt': _SET_SRI_OVERRIDES_6E5EF6CC,
    'vte': _SET_SRI_OVERRIDES_6E5EF6CC,
    'warp': _SET_SRI_OVERRIDES_6E5EF6CC,
    'wezterm': _SET_SRI_OVERRIDES_6E5EF6CC,
    'xterm': _SET_SRI_OVERRIDES_6E5EF6CC,
    'xterm.js': _SET_SRI_OVERRIDES_6E5EF6CC,
}

# For terminals: bobcat, kitty
_SET_SFZ_OVERRIDES_73E5B60D: dict[str, tuple[tuple[int, int], ...]] = {
    'zeroer': (
        (0x1f3fb, 0x1f3ff,),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
    ),
}

# For terminals: foot
_SET_SFZ_OVERRIDES_8DDFF0C4: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x1f3fb, 0x1f3ff,),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
    ),
}

SFZ_OVERRIDES: dict[str, dict[str, tuple[tuple[int, int], ...]]] = {
    'bobcat': _SET_SFZ_OVERRIDES_73E5B60D,
    'foot': _SET_SFZ_OVERRIDES_8DDFF0C4,
    'kitty': _SET_SFZ_OVERRIDES_73E5B60D,
}

# For terminals: iterm2
_SET_VS16_OVERRIDES_82FD018F: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x00023, 0x00023,),  # Number Sign
        (0x0002a, 0x0002a,),  # Asterisk
        (0x00030, 0x00039,),  # Digit Zero              ..Digit Nine
    ),
}

# For terminals: alacritty, apple_terminal, contour, mlterm, pterm, st, terminology, urxvt, vte, wezterm, xterm, xterm.js
_SET_VS16_OVERRIDES_F48531B2: dict[str, tuple[tuple[int, int], ...]] = {
    'narrower': (
        (0x00023, 0x00023,),  # Number Sign
        (0x0002a, 0x0002a,),  # Asterisk
        (0x00030, 0x00039,),  # Digit Zero              ..Digit Nine
        (0x000a9, 0x000a9,),  # Copyright Sign
        (0x000ae, 0x000ae,),  # Registered Sign
        (0x0203c, 0x0203c,),  # Double Exclamation Mark
        (0x02049, 0x02049,),  # Exclamation Question Mark
        (0x02122, 0x02122,),  # Trade Mark Sign
        (0x02139, 0x02139,),  # Information Source
        (0x02194, 0x02199,),  # Left Right Arrow        ..South West Arrow
        (0x021a9, 0x021aa,),  # Leftwards Arrow With Hoo..Rightwards Arrow With Ho
        (0x02328, 0x02328,),  # Keyboard
        (0x023cf, 0x023cf,),  # Eject Symbol
        (0x023ed, 0x023ef,),  # Black Right-pointing Dou..Black Right-pointing Tri
        (0x023f1, 0x023f2,),  # Stopwatch               ..Timer Clock
        (0x023f8, 0x023fa,),  # Double Vertical Bar     ..Black Circle For Record
        (0x024c2, 0x024c2,),  # Circled Latin Capital Letter M
        (0x025aa, 0x025ab,),  # Black Small Square      ..White Small Square
        (0x025b6, 0x025b6,),  # Black Right-pointing Triangle
        (0x025c0, 0x025c0,),  # Black Left-pointing Triangle
        (0x025fb, 0x025fc,),  # White Medium Square     ..Black Medium Square
        (0x02600, 0x02604,),  # Black Sun With Rays     ..Comet
        (0x0260e, 0x0260e,),  # Black Telephone
        (0x02611, 0x02611,),  # Ballot Box With Check
        (0x02618, 0x02618,),  # Shamrock
        (0x0261d, 0x0261d,),  # White Up Pointing Index
        (0x02620, 0x02620,),  # Skull And Crossbones
        (0x02622, 0x02623,),  # Radioactive Sign        ..Biohazard Sign
        (0x02626, 0x02626,),  # Orthodox Cross
        (0x0262a, 0x0262a,),  # Star And Crescent
        (0x0262e, 0x0262f,),  # Peace Symbol            ..Yin Yang
        (0x02638, 0x0263a,),  # Wheel Of Dharma         ..White Smiling Face
        (0x02640, 0x02640,),  # Female Sign
        (0x02642, 0x02642,),  # Male Sign
        (0x0265f, 0x02660,),  # Black Chess Pawn        ..Black Spade Suit
        (0x02663, 0x02663,),  # Black Club Suit
        (0x02665, 0x02666,),  # Black Heart Suit        ..Black Diamond Suit
        (0x02668, 0x02668,),  # Hot Springs
        (0x0267b, 0x0267b,),  # Black Universal Recycling Symbol
        (0x0267e, 0x0267e,),  # Permanent Paper Sign
        (0x02692, 0x02692,),  # Hammer And Pick
        (0x02694, 0x02697,),  # Crossed Swords          ..Alembic
        (0x02699, 0x02699,),  # Gear
        (0x0269b, 0x0269c,),  # Atom Symbol             ..Fleur-de-lis
        (0x026a0, 0x026a0,),  # Warning Sign
        (0x026a7, 0x026a7,),  # Male With Stroke And Male And Female Sign
        (0x026b0, 0x026b1,),  # Coffin                  ..Funeral Urn
        (0x026c8, 0x026c8,),  # Thunder Cloud And Rain
        (0x026cf, 0x026cf,),  # Pick
        (0x026d1, 0x026d1,),  # Helmet With White Cross
        (0x026d3, 0x026d3,),  # Chains
        (0x026e9, 0x026e9,),  # Shinto Shrine
        (0x026f0, 0x026f1,),  # Mountain                ..Umbrella On Ground
        (0x026f4, 0x026f4,),  # Ferry
        (0x026f7, 0x026f9,),  # Skier                   ..Person With Ball
        (0x02702, 0x02702,),  # Black Scissors
        (0x02708, 0x02709,),  # Airplane                ..Envelope
        (0x0270c, 0x0270d,),  # Victory Hand            ..Writing Hand
        (0x0270f, 0x0270f,),  # Pencil
        (0x02712, 0x02712,),  # Black Nib
        (0x02714, 0x02714,),  # Heavy Check Mark
        (0x02716, 0x02716,),  # Heavy Multiplication X
        (0x0271d, 0x0271d,),  # Latin Cross
        (0x02721, 0x02721,),  # Star Of David
        (0x02733, 0x02734,),  # Eight Spoked Asterisk   ..Eight Pointed Black Star
        (0x02744, 0x02744,),  # Snowflake
        (0x02747, 0x02747,),  # Sparkle
        (0x02763, 0x02764,),  # Heavy Heart Exclamation ..Heavy Black Heart
        (0x027a1, 0x027a1,),  # Black Rightwards Arrow
        (0x02934, 0x02935,),  # Arrow Pointing Rightward..Arrow Pointing Rightward
        (0x02b05, 0x02b07,),  # Leftwards Black Arrow   ..Downwards Black Arrow
        (0x1f170, 0x1f171,),  # Negative Squared Latin C..Negative Squared Latin C
        (0x1f17e, 0x1f17f,),  # Negative Squared Latin C..Negative Squared Latin C
        (0x1f321, 0x1f321,),  # Thermometer
        (0x1f324, 0x1f32c,),  # White Sun With Small Clo..Wind Blowing Face
        (0x1f336, 0x1f336,),  # Hot Pepper
        (0x1f37d, 0x1f37d,),  # Fork And Knife With Plate
        (0x1f396, 0x1f397,),  # Military Medal          ..Reminder Ribbon
        (0x1f399, 0x1f39b,),  # Studio Microphone       ..Control Knobs
        (0x1f39e, 0x1f39f,),  # Film Frames             ..Admission Tickets
        (0x1f3cb, 0x1f3ce,),  # Weight Lifter           ..Racing Car
        (0x1f3d4, 0x1f3df,),  # Snow Capped Mountain    ..Stadium
        (0x1f3f3, 0x1f3f3,),  # Waving White Flag
        (0x1f3f5, 0x1f3f5,),  # Rosette
        (0x1f3f7, 0x1f3f7,),  # Label
        (0x1f43f, 0x1f43f,),  # Chipmunk
        (0x1f441, 0x1f441,),  # Eye
        (0x1f4fd, 0x1f4fd,),  # Film Projector
        (0x1f549, 0x1f54a,),  # Om Symbol               ..Dove Of Peace
        (0x1f56f, 0x1f570,),  # Candle                  ..Mantelpiece Clock
        (0x1f573, 0x1f579,),  # Hole                    ..Joystick
        (0x1f587, 0x1f587,),  # Linked Paperclips
        (0x1f58a, 0x1f58d,),  # Lower Left Ballpoint Pen..Lower Left Crayon
        (0x1f590, 0x1f590,),  # Raised Hand With Fingers Splayed
        (0x1f5a5, 0x1f5a5,),  # Desktop Computer
        (0x1f5a8, 0x1f5a8,),  # Printer
        (0x1f5b1, 0x1f5b2,),  # Three Button Mouse      ..Trackball
        (0x1f5bc, 0x1f5bc,),  # Frame With Picture
        (0x1f5c2, 0x1f5c4,),  # Card Index Dividers     ..File Cabinet
        (0x1f5d1, 0x1f5d3,),  # Wastebasket             ..Spiral Calendar Pad
        (0x1f5dc, 0x1f5de,),  # Compression             ..Rolled-up Newspaper
        (0x1f5e1, 0x1f5e1,),  # Dagger Knife
        (0x1f5e3, 0x1f5e3,),  # Speaking Head In Silhouette
        (0x1f5e8, 0x1f5e8,),  # Left Speech Bubble
        (0x1f5ef, 0x1f5ef,),  # Right Anger Bubble
        (0x1f5f3, 0x1f5f3,),  # Ballot Box With Ballot
        (0x1f5fa, 0x1f5fa,),  # World Map
        (0x1f6cb, 0x1f6cb,),  # Couch And Lamp
        (0x1f6cd, 0x1f6cf,),  # Shopping Bags           ..Bed
        (0x1f6e0, 0x1f6e5,),  # Hammer And Wrench       ..Motor Boat
        (0x1f6e9, 0x1f6e9,),  # Small Airplane
        (0x1f6f0, 0x1f6f0,),  # Satellite
        (0x1f6f3, 0x1f6f3,),  # Passenger Ship
    ),
}

VS16_OVERRIDES: dict[str, dict[str, tuple[tuple[int, int], ...]]] = {
    'alacritty': _SET_VS16_OVERRIDES_F48531B2,
    'apple_terminal': _SET_VS16_OVERRIDES_F48531B2,
    'contour': _SET_VS16_OVERRIDES_F48531B2,
    'iterm2': _SET_VS16_OVERRIDES_82FD018F,
    'mlterm': _SET_VS16_OVERRIDES_F48531B2,
    'pterm': _SET_VS16_OVERRIDES_F48531B2,
    'st': _SET_VS16_OVERRIDES_F48531B2,
    'terminology': _SET_VS16_OVERRIDES_F48531B2,
    'urxvt': _SET_VS16_OVERRIDES_F48531B2,
    'vte': _SET_VS16_OVERRIDES_F48531B2,
    'wezterm': _SET_VS16_OVERRIDES_F48531B2,
    'xterm': _SET_VS16_OVERRIDES_F48531B2,
    'xterm.js': _SET_VS16_OVERRIDES_F48531B2,
}

# For terminals: rio
_SET_VS15_OVERRIDES_0F19A03D: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x023eb, 0x023ec,),  # Black Up-pointing Double..Black Down-pointing Doub
        (0x023f0, 0x023f0,),  # Alarm Clock
        (0x026ce, 0x026ce,),  # Ophiuchus
        (0x02705, 0x02705,),  # White Heavy Check Mark
        (0x0270a, 0x0270b,),  # Raised Fist             ..Raised Hand
        (0x02728, 0x02728,),  # Sparkles
        (0x0274c, 0x0274c,),  # Cross Mark
        (0x0274e, 0x0274e,),  # Negative Squared Cross Mark
        (0x02754, 0x02755,),  # White Question Mark Orna..White Exclamation Mark O
        (0x02795, 0x02797,),  # Heavy Plus Sign         ..Heavy Division Sign
        (0x027b0, 0x027b0,),  # Curly Loop
        (0x027bf, 0x027bf,),  # Double Curly Loop
    ),
}

# For terminals: bobcat
_SET_VS15_OVERRIDES_72B9888B: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x03030, 0x03030,),  # Wavy Dash
        (0x0303d, 0x0303d,),  # Part Alternation Mark
        (0x03297, 0x03297,),  # Circled Ideograph Congratulation
        (0x03299, 0x03299,),  # Circled Ideograph Secret
        (0x1f202, 0x1f202,),  # Squared Katakana Sa
        (0x1f237, 0x1f237,),  # Squared Cjk Unified Ideograph-6708
    ),
}

# For terminals: warp
_SET_VS15_OVERRIDES_D32653C3: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x0231a, 0x0231b,),  # Watch                   ..Hourglass
        (0x023e9, 0x023ec,),  # Black Right-pointing Dou..Black Down-pointing Doub
        (0x023f0, 0x023f0,),  # Alarm Clock
        (0x023f3, 0x023f3,),  # Hourglass With Flowing Sand
        (0x025fd, 0x025fe,),  # White Medium Small Squar..Black Medium Small Squar
        (0x02614, 0x02615,),  # Umbrella With Rain Drops..Hot Beverage
        (0x02648, 0x02653,),  # Aries                   ..Pisces
        (0x0267f, 0x0267f,),  # Wheelchair Symbol
        (0x02693, 0x02693,),  # Anchor
        (0x026a1, 0x026a1,),  # High Voltage Sign
        (0x026aa, 0x026ab,),  # Medium White Circle     ..Medium Black Circle
        (0x026bd, 0x026be,),  # Soccer Ball             ..Baseball
        (0x026c4, 0x026c5,),  # Snowman Without Snow    ..Sun Behind Cloud
        (0x026ce, 0x026ce,),  # Ophiuchus
        (0x026d4, 0x026d4,),  # No Entry
        (0x026ea, 0x026ea,),  # Church
        (0x026f2, 0x026f3,),  # Fountain                ..Flag In Hole
        (0x026f5, 0x026f5,),  # Sailboat
        (0x026fa, 0x026fa,),  # Tent
        (0x026fd, 0x026fd,),  # Fuel Pump
        (0x02705, 0x02705,),  # White Heavy Check Mark
        (0x0270a, 0x0270b,),  # Raised Fist             ..Raised Hand
        (0x02728, 0x02728,),  # Sparkles
        (0x0274c, 0x0274c,),  # Cross Mark
        (0x0274e, 0x0274e,),  # Negative Squared Cross Mark
        (0x02753, 0x02755,),  # Black Question Mark Orna..White Exclamation Mark O
        (0x02757, 0x02757,),  # Heavy Exclamation Mark Symbol
        (0x02795, 0x02797,),  # Heavy Plus Sign         ..Heavy Division Sign
        (0x027b0, 0x027b0,),  # Curly Loop
        (0x027bf, 0x027bf,),  # Double Curly Loop
        (0x02b1b, 0x02b1c,),  # Black Large Square      ..White Large Square
        (0x02b50, 0x02b50,),  # White Medium Star
        (0x02b55, 0x02b55,),  # Heavy Large Circle
        (0x1f004, 0x1f004,),  # Mahjong Tile Red Dragon
        (0x1f30d, 0x1f30f,),  # Earth Globe Europe-afric..Earth Globe Asia-austral
        (0x1f315, 0x1f315,),  # Full Moon Symbol
        (0x1f31c, 0x1f31c,),  # Last Quarter Moon With Face
        (0x1f378, 0x1f378,),  # Cocktail Glass
        (0x1f393, 0x1f393,),  # Graduation Cap
        (0x1f3a7, 0x1f3a7,),  # Headphone
        (0x1f3ac, 0x1f3ae,),  # Clapper Board           ..Video Game
        (0x1f3c2, 0x1f3c2,),  # Snowboarder
        (0x1f3c4, 0x1f3c4,),  # Surfer
        (0x1f3c6, 0x1f3c6,),  # Trophy
        (0x1f3ca, 0x1f3ca,),  # Swimmer
        (0x1f3e0, 0x1f3e0,),  # House Building
        (0x1f3ed, 0x1f3ed,),  # Factory
        (0x1f408, 0x1f408,),  # Cat
        (0x1f415, 0x1f415,),  # Dog
        (0x1f41f, 0x1f41f,),  # Fish
        (0x1f426, 0x1f426,),  # Bird
        (0x1f442, 0x1f442,),  # Ear
        (0x1f446, 0x1f449,),  # White Up Pointing Backha..White Right Pointing Bac
        (0x1f44d, 0x1f44e,),  # Thumbs Up Sign          ..Thumbs Down Sign
        (0x1f453, 0x1f453,),  # Eyeglasses
        (0x1f46a, 0x1f46a,),  # Family
        (0x1f47d, 0x1f47d,),  # Extraterrestrial Alien
        (0x1f4a3, 0x1f4a3,),  # Bomb
        (0x1f4b0, 0x1f4b0,),  # Money Bag
        (0x1f4b3, 0x1f4b3,),  # Credit Card
        (0x1f4bb, 0x1f4bb,),  # Personal Computer
        (0x1f4bf, 0x1f4bf,),  # Optical Disc
        (0x1f4cb, 0x1f4cb,),  # Clipboard
        (0x1f4da, 0x1f4da,),  # Books
        (0x1f4df, 0x1f4df,),  # Pager
        (0x1f4e4, 0x1f4e6,),  # Outbox Tray             ..Package
        (0x1f4ea, 0x1f4ed,),  # Closed Mailbox With Lowe..Open Mailbox With Lowere
        (0x1f4f7, 0x1f4f7,),  # Camera
        (0x1f4f9, 0x1f4fb,),  # Video Camera            ..Radio
        (0x1f508, 0x1f508,),  # Speaker
        (0x1f50d, 0x1f50d,),  # Left-pointing Magnifying Glass
        (0x1f512, 0x1f513,),  # Lock                    ..Open Lock
        (0x1f550, 0x1f567,),  # Clock Face One Oclock   ..Clock Face Twelve-thirty
        (0x1f610, 0x1f610,),  # Neutral Face
        (0x1f687, 0x1f687,),  # Metro
        (0x1f68d, 0x1f68d,),  # Oncoming Bus
        (0x1f691, 0x1f691,),  # Ambulance
        (0x1f694, 0x1f694,),  # Oncoming Police Car
        (0x1f698, 0x1f698,),  # Oncoming Automobile
        (0x1f6ad, 0x1f6ad,),  # No Smoking Symbol
        (0x1f6b2, 0x1f6b2,),  # Bicycle
        (0x1f6b9, 0x1f6ba,),  # Mens Symbol             ..Womens Symbol
        (0x1f6bc, 0x1f6bc,),  # Baby Symbol
    ),
}

# For terminals: alacritty, apple_terminal, contour, extraterm, iterm2, konsole, mlterm, pterm, st, terminology, urxvt, vte, wezterm, xterm, xterm.js
_SET_VS15_OVERRIDES_EC2D40AC: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x0231a, 0x0231b,),  # Watch                   ..Hourglass
        (0x023e9, 0x023ec,),  # Black Right-pointing Dou..Black Down-pointing Doub
        (0x023f0, 0x023f0,),  # Alarm Clock
        (0x023f3, 0x023f3,),  # Hourglass With Flowing Sand
        (0x025fd, 0x025fe,),  # White Medium Small Squar..Black Medium Small Squar
        (0x02614, 0x02615,),  # Umbrella With Rain Drops..Hot Beverage
        (0x02648, 0x02653,),  # Aries                   ..Pisces
        (0x0267f, 0x0267f,),  # Wheelchair Symbol
        (0x02693, 0x02693,),  # Anchor
        (0x026a1, 0x026a1,),  # High Voltage Sign
        (0x026aa, 0x026ab,),  # Medium White Circle     ..Medium Black Circle
        (0x026bd, 0x026be,),  # Soccer Ball             ..Baseball
        (0x026c4, 0x026c5,),  # Snowman Without Snow    ..Sun Behind Cloud
        (0x026ce, 0x026ce,),  # Ophiuchus
        (0x026d4, 0x026d4,),  # No Entry
        (0x026ea, 0x026ea,),  # Church
        (0x026f2, 0x026f3,),  # Fountain                ..Flag In Hole
        (0x026f5, 0x026f5,),  # Sailboat
        (0x026fa, 0x026fa,),  # Tent
        (0x026fd, 0x026fd,),  # Fuel Pump
        (0x02705, 0x02705,),  # White Heavy Check Mark
        (0x0270a, 0x0270b,),  # Raised Fist             ..Raised Hand
        (0x02728, 0x02728,),  # Sparkles
        (0x0274c, 0x0274c,),  # Cross Mark
        (0x0274e, 0x0274e,),  # Negative Squared Cross Mark
        (0x02753, 0x02755,),  # Black Question Mark Orna..White Exclamation Mark O
        (0x02757, 0x02757,),  # Heavy Exclamation Mark Symbol
        (0x02795, 0x02797,),  # Heavy Plus Sign         ..Heavy Division Sign
        (0x027b0, 0x027b0,),  # Curly Loop
        (0x027bf, 0x027bf,),  # Double Curly Loop
        (0x02b1b, 0x02b1c,),  # Black Large Square      ..White Large Square
        (0x02b50, 0x02b50,),  # White Medium Star
        (0x02b55, 0x02b55,),  # Heavy Large Circle
        (0x03030, 0x03030,),  # Wavy Dash
        (0x0303d, 0x0303d,),  # Part Alternation Mark
        (0x03297, 0x03297,),  # Circled Ideograph Congratulation
        (0x03299, 0x03299,),  # Circled Ideograph Secret
        (0x1f004, 0x1f004,),  # Mahjong Tile Red Dragon
        (0x1f202, 0x1f202,),  # Squared Katakana Sa
        (0x1f21a, 0x1f21a,),  # Squared Cjk Unified Ideograph-7121
        (0x1f22f, 0x1f22f,),  # Squared Cjk Unified Ideograph-6307
        (0x1f237, 0x1f237,),  # Squared Cjk Unified Ideograph-6708
        (0x1f30d, 0x1f30f,),  # Earth Globe Europe-afric..Earth Globe Asia-austral
        (0x1f315, 0x1f315,),  # Full Moon Symbol
        (0x1f31c, 0x1f31c,),  # Last Quarter Moon With Face
        (0x1f378, 0x1f378,),  # Cocktail Glass
        (0x1f393, 0x1f393,),  # Graduation Cap
        (0x1f3a7, 0x1f3a7,),  # Headphone
        (0x1f3ac, 0x1f3ae,),  # Clapper Board           ..Video Game
        (0x1f3c2, 0x1f3c2,),  # Snowboarder
        (0x1f3c4, 0x1f3c4,),  # Surfer
        (0x1f3c6, 0x1f3c6,),  # Trophy
        (0x1f3ca, 0x1f3ca,),  # Swimmer
        (0x1f3e0, 0x1f3e0,),  # House Building
        (0x1f3ed, 0x1f3ed,),  # Factory
        (0x1f408, 0x1f408,),  # Cat
        (0x1f415, 0x1f415,),  # Dog
        (0x1f41f, 0x1f41f,),  # Fish
        (0x1f426, 0x1f426,),  # Bird
        (0x1f442, 0x1f442,),  # Ear
        (0x1f446, 0x1f449,),  # White Up Pointing Backha..White Right Pointing Bac
        (0x1f44d, 0x1f44e,),  # Thumbs Up Sign          ..Thumbs Down Sign
        (0x1f453, 0x1f453,),  # Eyeglasses
        (0x1f46a, 0x1f46a,),  # Family
        (0x1f47d, 0x1f47d,),  # Extraterrestrial Alien
        (0x1f4a3, 0x1f4a3,),  # Bomb
        (0x1f4b0, 0x1f4b0,),  # Money Bag
        (0x1f4b3, 0x1f4b3,),  # Credit Card
        (0x1f4bb, 0x1f4bb,),  # Personal Computer
        (0x1f4bf, 0x1f4bf,),  # Optical Disc
        (0x1f4cb, 0x1f4cb,),  # Clipboard
        (0x1f4da, 0x1f4da,),  # Books
        (0x1f4df, 0x1f4df,),  # Pager
        (0x1f4e4, 0x1f4e6,),  # Outbox Tray             ..Package
        (0x1f4ea, 0x1f4ed,),  # Closed Mailbox With Lowe..Open Mailbox With Lowere
        (0x1f4f7, 0x1f4f7,),  # Camera
        (0x1f4f9, 0x1f4fb,),  # Video Camera            ..Radio
        (0x1f508, 0x1f508,),  # Speaker
        (0x1f50d, 0x1f50d,),  # Left-pointing Magnifying Glass
        (0x1f512, 0x1f513,),  # Lock                    ..Open Lock
        (0x1f550, 0x1f567,),  # Clock Face One Oclock   ..Clock Face Twelve-thirty
        (0x1f610, 0x1f610,),  # Neutral Face
        (0x1f687, 0x1f687,),  # Metro
        (0x1f68d, 0x1f68d,),  # Oncoming Bus
        (0x1f691, 0x1f691,),  # Ambulance
        (0x1f694, 0x1f694,),  # Oncoming Police Car
        (0x1f698, 0x1f698,),  # Oncoming Automobile
        (0x1f6ad, 0x1f6ad,),  # No Smoking Symbol
        (0x1f6b2, 0x1f6b2,),  # Bicycle
        (0x1f6b9, 0x1f6ba,),  # Mens Symbol             ..Womens Symbol
        (0x1f6bc, 0x1f6bc,),  # Baby Symbol
    ),
}

VS15_OVERRIDES: dict[str, dict[str, tuple[tuple[int, int], ...]]] = {
    'alacritty': _SET_VS15_OVERRIDES_EC2D40AC,
    'apple_terminal': _SET_VS15_OVERRIDES_EC2D40AC,
    'bobcat': _SET_VS15_OVERRIDES_72B9888B,
    'contour': _SET_VS15_OVERRIDES_EC2D40AC,
    'extraterm': _SET_VS15_OVERRIDES_EC2D40AC,
    'iterm2': _SET_VS15_OVERRIDES_EC2D40AC,
    'konsole': _SET_VS15_OVERRIDES_EC2D40AC,
    'mlterm': _SET_VS15_OVERRIDES_EC2D40AC,
    'pterm': _SET_VS15_OVERRIDES_EC2D40AC,
    'rio': _SET_VS15_OVERRIDES_0F19A03D,
    'st': _SET_VS15_OVERRIDES_EC2D40AC,
    'terminology': _SET_VS15_OVERRIDES_EC2D40AC,
    'urxvt': _SET_VS15_OVERRIDES_EC2D40AC,
    'vte': _SET_VS15_OVERRIDES_EC2D40AC,
    'warp': _SET_VS15_OVERRIDES_D32653C3,
    'wezterm': _SET_VS15_OVERRIDES_EC2D40AC,
    'xterm': _SET_VS15_OVERRIDES_EC2D40AC,
    'xterm.js': _SET_VS15_OVERRIDES_EC2D40AC,
}

# For terminals: alacritty
_SET_NARROW_OVERRIDES_19FF3EBB: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x017a4, 0x017a4,),  # Khmer Independent Vowel Qaa
        (0x017d8, 0x017d8,),  # Khmer Sign Beyyal
    ),
    'narrow_zeroer': (
        (0x000ad, 0x000ad,),  # Soft Hyphen
        (0x00605, 0x00605,),  # Arabic Number Mark Above
        (0x0070f, 0x0070f,),  # Syriac Abbreviation Mark
        (0x00890, 0x00891,),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x00d4e, 0x00d4e,),  # Malayalam Letter Dot Reph
        (0x0a8fa, 0x0a8fa,),  # Devanagari Caret
        (0x0ff9e, 0x0ff9f,),  # Halfwidth Katakana Voice..Halfwidth Katakana Semi-
        (0x111c2, 0x111c3,),  # Sharada Sign Jihvamuliya..Sharada Sign Upadhmaniya
        (0x1193f, 0x1193f,),  # Dives Akuru Prefixed Nasal Sign
        (0x11941, 0x11941,),  # Dives Akuru Initial Ra
        (0x11a84, 0x11a89,),  # Soyombo Sign Jihvamuliya..Soyombo Cluster-initial
        (0x11d46, 0x11d46,),  # Masaram Gondi Repha
        (0x11f02, 0x11f02,),  # Kawi Sign Repha
    ),
}

# For terminals: konsole, wezterm
_SET_NARROW_OVERRIDES_3E5607A0: dict[str, tuple[tuple[int, int], ...]] = {
    'narrow_zeroer': (
        (0x000ad, 0x000ad,),  # Soft Hyphen
        (0x00600, 0x00605,),  # Arabic Number Sign      ..Arabic Number Mark Above
        (0x006dd, 0x006dd,),  # Arabic End Of Ayah
        (0x0070f, 0x0070f,),  # Syriac Abbreviation Mark
        (0x00890, 0x00891,),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x110bd, 0x110bd,),  # Kaithi Number Sign
        (0x110cd, 0x110cd,),  # Kaithi Number Sign Above
    ),
}

# For terminals: kitty
_SET_NARROW_OVERRIDES_4F7FBF42: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x0261d, 0x0261d,),  # White Up Pointing Index
        (0x026f9, 0x026f9,),  # Person With Ball
        (0x0270c, 0x0270d,),  # Victory Hand            ..Writing Hand
        (0x1f3cb, 0x1f3cc,),  # Weight Lifter           ..Golfer
        (0x1f574, 0x1f575,),  # Man In Business Suit Lev..Sleuth Or Spy
        (0x1f590, 0x1f590,),  # Raised Hand With Fingers Splayed
    ),
    'narrow_zeroer': (
        (0x000ad, 0x000ad,),  # Soft Hyphen
        (0x00600, 0x00605,),  # Arabic Number Sign      ..Arabic Number Mark Above
        (0x006dd, 0x006dd,),  # Arabic End Of Ayah
        (0x0070f, 0x0070f,),  # Syriac Abbreviation Mark
        (0x00890, 0x00891,),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x0ff9e, 0x0ff9f,),  # Halfwidth Katakana Voice..Halfwidth Katakana Semi-
        (0x110bd, 0x110bd,),  # Kaithi Number Sign
        (0x110cd, 0x110cd,),  # Kaithi Number Sign Above
    ),
}

# For terminals: xterm.js
_SET_NARROW_OVERRIDES_763C1244: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x1f93b, 0x1f93b,),  # Modern Pentathlon
        (0x1f946, 0x1f946,),  # Rifle
    ),
    'narrow_zeroer': (
        (0x00600, 0x00605,),  # Arabic Number Sign      ..Arabic Number Mark Above
        (0x006dd, 0x006dd,),  # Arabic End Of Ayah
        (0x0070f, 0x0070f,),  # Syriac Abbreviation Mark
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x110bd, 0x110bd,),  # Kaithi Number Sign
        (0x110cd, 0x110cd,),  # Kaithi Number Sign Above
    ),
}

# For terminals: foot, pterm, rio, vte
_SET_NARROW_OVERRIDES_8B501630: dict[str, tuple[tuple[int, int], ...]] = {
    'narrow_zeroer': (
        (0x00600, 0x00605,),  # Arabic Number Sign      ..Arabic Number Mark Above
        (0x006dd, 0x006dd,),  # Arabic End Of Ayah
        (0x0070f, 0x0070f,),  # Syriac Abbreviation Mark
        (0x00890, 0x00891,),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x110bd, 0x110bd,),  # Kaithi Number Sign
        (0x110cd, 0x110cd,),  # Kaithi Number Sign Above
    ),
}

# For terminals: st, urxvt, xterm
_SET_NARROW_OVERRIDES_9BC01E8D: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x03248, 0x0324f,),  # Circled Number Ten On Bl..Circled Number Eighty On
    ),
}

# For terminals: bobcat
_SET_NARROW_OVERRIDES_B19CC1BB: dict[str, tuple[tuple[int, int], ...]] = {
    'narrow_zeroer': (
        (0x00890, 0x00891,),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x110bd, 0x110bd,),  # Kaithi Number Sign
        (0x110cd, 0x110cd,),  # Kaithi Number Sign Above
    ),
}

# For terminals: iterm2
_SET_NARROW_OVERRIDES_C5500A15: dict[str, tuple[tuple[int, int], ...]] = {
    'narrow_zeroer': (
        (0x000ad, 0x000ad,),  # Soft Hyphen
    ),
}

# For terminals: warp
_SET_NARROW_OVERRIDES_FDEEB1C7: dict[str, tuple[tuple[int, int], ...]] = {
    'wider': (
        (0x017a4, 0x017a4,),  # Khmer Independent Vowel Qaa
        (0x017d8, 0x017d8,),  # Khmer Sign Beyyal
    ),
    'narrow_zeroer': (
        (0x000ad, 0x000ad,),  # Soft Hyphen
        (0x00605, 0x00605,),  # Arabic Number Mark Above
        (0x0070f, 0x0070f,),  # Syriac Abbreviation Mark
        (0x00890, 0x00891,),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (0x008e2, 0x008e2,),  # Arabic Disputed End Of Ayah
        (0x00d4e, 0x00d4e,),  # Malayalam Letter Dot Reph
        (0x0a8fa, 0x0a8fa,),  # Devanagari Caret
        (0x0ff9e, 0x0ff9f,),  # Halfwidth Katakana Voice..Halfwidth Katakana Semi-
        (0x111c2, 0x111c3,),  # Sharada Sign Jihvamuliya..Sharada Sign Upadhmaniya
        (0x1193f, 0x1193f,),  # Dives Akuru Prefixed Nasal Sign
        (0x11941, 0x11941,),  # Dives Akuru Initial Ra
        (0x11a3a, 0x11a3a,),  # Zanabazar Square Cluster-initial Letter Ra
        (0x11a84, 0x11a89,),  # Soyombo Sign Jihvamuliya..Soyombo Cluster-initial
        (0x11d46, 0x11d46,),  # Masaram Gondi Repha
        (0x11f02, 0x11f02,),  # Kawi Sign Repha
    ),
}

NARROW_OVERRIDES: dict[str, dict[str, tuple[tuple[int, int], ...]]] = {
    'alacritty': _SET_NARROW_OVERRIDES_19FF3EBB,
    'bobcat': _SET_NARROW_OVERRIDES_B19CC1BB,
    'foot': _SET_NARROW_OVERRIDES_8B501630,
    'iterm2': _SET_NARROW_OVERRIDES_C5500A15,
    'kitty': _SET_NARROW_OVERRIDES_4F7FBF42,
    'konsole': _SET_NARROW_OVERRIDES_3E5607A0,
    'pterm': _SET_NARROW_OVERRIDES_8B501630,
    'rio': _SET_NARROW_OVERRIDES_8B501630,
    'st': _SET_NARROW_OVERRIDES_9BC01E8D,
    'urxvt': _SET_NARROW_OVERRIDES_9BC01E8D,
    'vte': _SET_NARROW_OVERRIDES_8B501630,
    'warp': _SET_NARROW_OVERRIDES_FDEEB1C7,
    'wezterm': _SET_NARROW_OVERRIDES_3E5607A0,
    'xterm': _SET_NARROW_OVERRIDES_9BC01E8D,
    'xterm.js': _SET_NARROW_OVERRIDES_763C1244,
}

__all__ = ['WIDE_OVERRIDES', 'SRI_OVERRIDES', 'SFZ_OVERRIDES', 'VS16_OVERRIDES', 'VS15_OVERRIDES', 'NARROW_OVERRIDES']



"""
OSC 8 hyperlink parsing and measurement.

.. versionadded:: 0.7.0
"""

from __future__ import annotations

# std imports
import re

import typing

# local
from ._width import width as _width
from .escape_sequences import _SEQUENCE_CLASSIFY

HYPERLINK_OPEN_RE = re.compile(r'\x1b]8;([^;]*);([^\x07\x1b]*)(\x07|\x1b\\)')
HYPERLINK_CLOSE_RE = re.compile(r'\x1b]8;;(\x07|\x1b\\)')


class HyperlinkParams(typing.NamedTuple):
    r"""
    Parsed parameters from an OSC 8 hyperlink open sequence.

    :param url: The hyperlink URL.
    :param params: Colon-separated metadata string (often empty).
    :param terminator: Sequence terminator (``\x07`` or ``\x1b\\``).
    """

    url: str
    params: str = ''
    terminator: str = '\x07'

    @classmethod
    def parse(cls, seq: str) -> HyperlinkParams | None:
        r"""
        Parse an OSC 8 open sequence string.

        Returns ``None`` if *seq* is not a valid OSC 8 open.

        Example::

            >>> HyperlinkParams.parse('\x1b]8;;http://example.com\x07')
            HyperlinkParams(url='http://example.com', params='', terminator='\\x07')
        """
        m = HYPERLINK_OPEN_RE.match(seq)
        if m is None:
            return None
        return cls(url=m.group(2), params=m.group(1), terminator=m.group(3))

    def make_open(self) -> str:
        """Generate the OSC 8 open escape sequence."""
        return f'\x1b]8;{self.params};{self.url}{self.terminator}'

    def make_close(self) -> str:
        """Generate the OSC 8 close escape sequence."""
        return f'\x1b]8;;{self.terminator}'


class Hyperlink(typing.NamedTuple):
    """
    A complete OSC 8 hyperlink with target and inner text.

    :param params: Parsed open sequence parameters.
    :param text: Inner text between the open and close sequences.
    """

    params: HyperlinkParams
    text: str

    @classmethod
    def find_close(cls, text: str, open_end: int) -> tuple[int, int]:
        """
        Find the matching OSC 8 close sequence.

        Searches 'text' starting at 'open_end', the position just past the open
        sequence.  Returns position of close sequence ``(close_start,
        close_end)`` or ``(-1, -1)`` if not found.

        Per the OSC 8 specification, terminal emulators treat hyperlinks as a
        state attribute, not as nested HTML anchors.  A close sequence closes
        the current hyperlink regardless of how many open sequences preceded it.
        """
        m = HYPERLINK_CLOSE_RE.search(text, open_end)
        if m is None:
            return (-1, -1)
        return (m.start(), m.end())

    def display_width(
        self,
        *,
        control_codes: typing.Literal['parse', 'strict', 'ignore'] = 'parse',
        tabsize: int = 8,
        ambiguous_width: int = 1,
    ) -> int:
        r"""
        Measure the display width of the hyperlink's inner text.

        Delegates to :func:`wcwidth.width` with the given parameters.

        Example::

            >>> hl = Hyperlink.parse('\x1b]8;;http://example.com\x07Hello\x1b]8;;\x07', 0)
            >>> hl.display_width()
            5
        """
        return _width(
            self.text,
            control_codes=control_codes,
            tabsize=tabsize,
            ambiguous_width=ambiguous_width,
        )

    @classmethod
    def parse(cls, text: str, start: int = 0) -> Hyperlink | None:
        r"""
        Parse a complete OSC 8 hyperlink unit from *text* at position *start*.

        Locates the open sequence, finds the matching close, and returns a
        ``Hyperlink`` containing the parsed parameters and inner text.  Returns
        ``None`` if the text at *start* is not a complete OSC 8 hyperlink.

        Example::

            >>> Hyperlink.parse('\x1b]8;;http://example.com\x07Hello\x1b]8;;\x07')
            Hyperlink(params=HyperlinkParams(url='http://example.com', ...), text='Hello')
        """
        m = _SEQUENCE_CLASSIFY.match(text, start)
        if m is None:
            return None
        params = HyperlinkParams.parse(m.group())
        if params is None:
            return None
        close_start, close_end = cls.find_close(text, m.end())
        if (close_start, close_end) == (-1, -1):
            return None
        return cls(params=params, text=text[m.end():close_start])

    def make_sequence(self) -> str:
        """Rebuild the complete OSC 8 hyperlink escape sequence."""
        return self.params.make_open() + self.text + self.params.make_close()
